% Programa MATLAB 5.3 - DEMOCRK.M
% Programa para simular uma rede de Kohonen aplicada no problema
% de classificacao de imagens. Neste caso as imagens sao 16 padroes
% com os caracteres hexadecimais 0-9,A-F. Cada imagem tem 7x7 pixels.
% Os padroes de treinamento podem ser com ou sem ruido sobreposto.
% 
% Durante o treinamento:
% - e' calculado o numero de clusters sobrepostas, usando para tal
%   as 16 imagens originais, isto e', sem ruido.
% - o tamanho da vizinhanca ao redor da unidade vencedora e o valor
%   da "learning rate" neta decrescem exponencialmente.
%
% Rotinas auxiliares: KHOUT.M, KHTRAIN.M, ORD.M, FLIP.M, HEXAPAT.M
%
% Autores: Cairo Nascimento (cairo@ita.br), Eduardo Botter e Joao Maluf Jr. (ITA)
% Data:  4/Nov/1997 - Verso 1.0 - MATLAB 4.2
%       11/Fev/2002 - Verso 1.1 - Atualizao para MATLAB 5.3

% Preparacao para posicionamento das figuras
clear all; close all; Nf=1;
figure(Nf);set(gcf,'Position',[  5  305   395   225]);
set(gcf,'DoubleBuffer','on'); % habilitar tracado mais rapido das linhas
                              % para mostrar animacoes em maquinas mais rapidas
figure(Nf+1);set(gcf,'Position',[405  305   395   225]);
set(gcf,'DoubleBuffer','on');

% Definicao dos padroes de treinamento e demais parametros
 hexapat;  % define os 16 padroes hexadecimais usados para
           % treinamento da rede
 % INPD: cada padrao e' uma coluna; INPD(i,j) = 1 ou -1.
 [Ninp,Npat]=size(INPD); % 
 % Definicao da string que identifica cada um dos padroes
 % de treinamento sem ruido
 HexStr=['0123456789ABCDEF']; % Padroes de treinamento: 0-9,A-F

% Definicao dos parametros de treinamento
 N1=10; N2=10;            % tamanho da grade
 Niter=500;            % numero de iteracoes
 neta0=0.5;             % constante de aprendizado inicial
 dist0=1+max(N1,N2)/2;  % distancia maxima inicial
 % neta e dist decaem com constante de tempo Niter/5
 % neta=neta0*exp(-5*iter/Niter)
 % dist=dist0*exp(-5*iter/Niter), dist=fix(dist),

 probflip=5/100;  % 1 = 100%, probabilidade de cada
 %probflip=0;
 nord=1;          % se nord > 0, padroes de treinamento sao
 % reordenados aleatoriamente em cada iteracao de treinamento

 % Npv = Numero de passos de treinamento entre visualizacoes
 % Sugestao para Npv: Niter/50
 Npv=10;
 
% Passos:
% 1 - Initialization: geracao aleatoria dos pesos da rede neural
% 2 - Training: treinamento da rede
%  2.1 - Sampling: escolher uma entrada x
%  2.2 - Similarity Matching: determinar a unidade vencedora
%        para esta entrada x, pelo criterio da menor distancia
%        [iv,jv] t.q. ||x-w(iv,jv)|| = min(||x-w(i,j)||) p/ todo i e j
%  2.3 - Updating: Ajustar os pesos de todas as unidades vizinhas
%        (no espaco i,j) da unidade vencedora
%           Dw = neta*(x-w)
%        Diminuir neta e o tamanho da vizinhanca com o treinamento
% Loop no passo 2

% Passo 1 - Geracao das matrizes de pesos
% ---------------------------------------
Seed0=4;             % semente inicial
%rand('state',Seed0);
rand('state',sum(100*clock));
W=2*(rand([Ninp,N1*N2])-0.5);

 % Visualizacao:
 % Construo da grade de sada
 figure(Nf);
 for i=1:N2;  % linhas horizontais
     plot([1 N1],[i i],'b-'); hold on;
 end;
 for j=1:N1; % linhas verticais
     plot([j j],[1 N2],'b-'); hold on;
 end;
 set(gca,'XTick',[1:N1]);
 set(gca,'YTick',[1:N2]);
 axis([0 N1+1 0 N2+1]);
 % Fazer a calibracao apontando unidade vencedora para cada padrao
 % de entrada sem ruido.
 Wm=zeros([N2,N1]);
 % Wm = matriz auxiliar usada descobrir numero de unidades respondendo
 % por mais de 1 padrao
 for pat=1:Npat;
     [iv,jv]=khout(W,INPD(:,pat),N1,N2);
     ht(pat)=text(jv+0.1,iv,HexStr(pat));
     % armazenar o numero da unidade vencedora
     Wm(iv,jv)=Wm(iv,jv)+1;
 end;
 set(ht,'Color',[1 0 0]); % amarelo
 set(ht,'VerticalAlignment','baseline');
 Ncs=Npat-sum(sum(Wm>0));  iter=0; ic=1; Lc(ic,:)=[iter Ncs];
 title(['Iterao ' num2str(iter) ' / ' num2str(Niter)]);
 xlabel(['Ncs = ' num2str(Ncs)]); 
 drawnow; refresh;
 disp('Pausing ...'); pause(1); disp('Out of pause');

% Passo 2 -  Loop das iteracoes de treinamento
% --------------------------------------------
INPD2=INPD; % matriz auxiliar
for iter=1:Niter;

    % Reordenar os padroes de treinamento
    ordpat=[1:Npat];
    if nord > 1; [ordpat2,ordpat3]=ord(ordpat); ordpat=ordpat3; end;

    % Adicionar ruido do tipo flip-flop nos padroes de entrada
    % usados para treinar a rede
    if 0 < probflip & probflip <= 1;
       INPD2=flip(INPD,probflip);
    end;
    
    % Loop para cada padrao de treinamento
    neta=neta0*exp(-5*iter/Niter);
    dist=round(0.35+dist0*exp(-5*iter/Niter));
    for pat=1:Npat;
        % Escolher uma entrada X
        X=INPD2(:,ordpat(pat));
        % Determinar a unidade vencedora para esta entrada X
        [iv,jv]=khout(W,X,N1,N2);
        % Ajustar os pesos de todas as unidades vizinhas
        % (no espaco i,j) da unidade vencedora
        W=khtrain(W,X,iv,jv,dist,neta,N1,N2);
    end;
        
    % Visualizacao:
    % Fazer a calibracao apontando unidade vencedora para cada padrao
    % de entrada sem ruido.
    if iter == Npv*fix(iter/Npv);
       Wm=zeros([N2,N1]);
       % Wm = matriz auxiliar usada descobrir numero de unidades
       % respondendo por mais de 1 padrao
       figure(Nf);
       delete(ht);
       for pat=1:Npat;
           [iv,jv]=khout(W,INPD(:,pat),N1,N2);
           ht(pat)=text(jv+0.1,iv,HexStr(pat));
           UV(pat,:)=[iv,jv];
           % armazenar o numero da unidade vencedora
           Wm(iv,jv)=Wm(iv,jv)+1;
       end;
       set(ht,'Color',[1 0 0]); % amarelo
       set(ht,'VerticalAlignment','baseline');
       Ncs=Npat-sum(sum(Wm>0)); ic=ic+1; Lc(ic,:)=[iter Ncs];
       title(['Iterao ' num2str(iter) ' / ' num2str(Niter) ' = ' num2str(fix(100*iter/Niter)) '%']);
       xlabel(['j, Ncs = ' num2str(Ncs), ', dist = ' num2str(dist)]);
       ylabel('i');
       drawnow; refresh;
       % plotar numero de clusters sobrepostas
       figure(Nf+1);
       plot(Lc(:,1),Lc(:,2),Lc(:,1),Lc(:,2),'r*');
       axis([0 inf 0 max(Lc(:,2))+1]); xlabel('No. de iteraes');
       title('No. de Clusters Sobrepostas'); grid;
       drawnow;
    end; % end "if iter == Npv*fix(iter/Npv)"

end   % for iter=1:Niter;

% Finalizacao grafica
figure(Nf);   hold off;
figure(Nf+1); hold off;

% EOF khhex.m
