function W=khtrain(W,X,iv,jv,dist,neta,N1,N2)
% Funcao para treinar os pesos de uma rede de Kohonen
% cujas unidades estao distribuidas em uma grade planar.
% O treinamento e' realizado apenas para as unidades
% na vizinhanca da unidade vencedora (Dw = neta*(x-w)).
% Assume-se:
%   N1 unidades no eixo horizontal,
%   N2 unidades no eixo vertical,
%   dim W = Ninp,N1*N2,
%   dim X = Ninp,1.
% Obs:
%   - Para calcular iv e jv use a funcao KHOUT.M.
%   - No caso das unidades da rede de Kohonen estarem
%     arranjadas na forma de linha, faca N2=1.
%
% Entradas:
% W = matriz de pesos da rede de Kohonen.
%    As unidades sao numeradas de baixo para cima, da esquerda
%    para direita. Total de unidades: N1*N2.
%    Coluna k de W corresponde aos pesos recebidos pela unidade
%    de numero k, onde 1 <= k <= N1*N2.
% X = vetor coluna com o padrao de entrada.
% iv,jv = posicao da unidade vencedora na grade N1 x N2.
% dist = distancia maxima da unidade vencedora, parametro usado
%    para definir a vizinhanca da unidade vencedora.
%    Assim todas as unidades tal que:
%       abs(i-iv) <= dist E abs(j-jv) <= dist
%    terao seus pesos atualizados. dist > 0.
% neta = constante de aprendizado. neta >= 0.
% N1, N2 = parametros que especificam o tamanho da grade de unidades.
%
% Saidas:
% W = matriz de pesos da rede de Kohonen, depois da atualizacao.
%
% Data: 3/nov/1997   -   Verso 1.0
% Autores: Cairo e Botter (ITA)

% Testar validade dos dados de entrada da funcao
[Ninp1,Nunid]=size(W);
[Ninp2,Ncol]=size(X);

if Ninp1 ~= Ninp2 | Nunid ~= N1*N2 | Ncol ~= 1;
   error('KHTRAIN: Verifique as dimensoes de W e X');
end;
if iv < 1 | iv > N2 | jv < 1 | jv > N1;
   error('KHTRAIN: iv e/ou jv fora do intervalo [1,N2],[1,N1]');
end;

if neta <0 | dist < 0 | dist ~= fix(dist);
   error('KHTRAIN: Verifique se neta >= 0 E dist > 0 E dist=numero inteiro');
end;

Ninp=Ninp1;

% Calcular limites inferiores e superiores que definem
% a regiao da vizinhanca em torno da unidade vencedora (iv,jv).
iv1=iv-dist; iv2=iv+dist;
jv1=jv-dist; jv2=jv+dist;
   
% Testar limites min e max dos indices da vizinhanca
% N1 = numero maximo de colunas
% N2 = numero maximo de linhas*
if iv1<1; iv1=1; end; if iv2>N2; iv2=N2; end;
if jv1<1; jv1=1; end; if jv2>N1; jv2=N1; end;

% Ajustar os pesos das unidades dentro da vizinhanca
% X1 = matriz com vetor X em cada coluna, usada
% para calculo de D(Wp) = Wp + neta*(X-Wp).
% Wp = colunas de W que vao ser alteradas
for j=1:jv2-jv1+1; X1(:,j)=X; end;
for i=iv1:iv2;
    % Iunid = vetor com os numeros das unidades na linha i
    % que devem ser atualizadas
    Iunid=(i-1)*N1+(jv1:jv2);
    % Atualizacao
    W(:,Iunid)=W(:,Iunid)+neta*(X1-W(:,Iunid));
end;

% EOF khtrain.m
