% MATLAB script 5.3 - file: KOH1.M
% Demonstracao da Rede de Kohonen.
% Exemplos das pags. 160-165 e 159 (figs. 12.6 a 12.8) do livro IA_CONTAUT,
% Cairo e Takashi, 2000.
% Dimensao do espaco de entrada = Ninp  >= 2
% Dimensao do espaco de sada   = Nout =  1 ou 2
% Rotinas de inicializacao dos pesos e plotagem
% para caso Ninp = 2
% Entradas distribuidas de acordo com a variavel Caso
% Unidades da rede neural distribuidas em uma grade
% com N1 unidades no eixo x1 (horizontal)
%   e N2 unidades no eixo x2 (eixo vertical)
% N1>=1 e N2>=1 (o caso N1=N2=1 nao e' valido)
% Obs: se N1=1 ou N2=1, teremos uma linha de neuronios
%       ao inves de uma grade.
%      Se N1<N2, os valores de N1 e N2 sao trocados entre si
%       pelo programa para facilitar alguns testes.
% Autor: Cairo L. Nascimento Jr.(ITA), cairo@ita.br
% Data: 3/Maio/1998, Revisao em 7/Fev/2002

% Preparacao do ambiente MATLAB
clear all; close all;; 
Nf=1; % Nf = numero da primeira figura
figure(Nf); set(gcf,'DoubleBuffer','on');
clf;

% Dados de entrada
x1min=-1; x1max=1;   % espaco de entrada = regiao dos dados
x2min=-1; x2max=1;   % de treinamento
N1=10; N2=10;        % numero de unidades nos eixos horizontal e vertical
                     % para caso linha de neuronios faca N1 =1 ou N2 = 1
Caso1=menu('Escolha o caso:', ...
          'Caso 1 = Unidades em linha', ...
          'Caso 2 = Unidades em grade');
if Caso1 ==1; N1=50; N2=1; end;
Niter=100;           % numero de iteracoes para treinamento
Npiter=50;           % numero de pontos em cada iteracao de treinamento
Npv=1;               % Numero de iteracoes entre visualizacoes

netamin=0.1;              % neta minimo durante aprendizado
neta0=0.6;                % taxa de aprendizado inicial
Kexp=5;                   % constante usada no calculo da vizinhanca
Rv0=sqrt(2)*max(N1,N2);   % raio inicial da vizinhanca de treinamento
if N1==1 | N2==1;
   Rv0=max(N1,N2);
end;
Seed0=4;                  % semente inicial

% Teste de validade de N1 e N2
if N1==1 & N2==1; error('N1=1 e N2=1. Escolha no vlida.'); end;
% No caso onde N2 > N1, troque os valores tal que N1 > N2
if N1 < N2; Nt=N1; N1=N2; N2=Nt; end;
% Set the initial seed
rand('seed',Seed0);

% Escolha da distribuicao do sinal de entrada x=[x1,x2]
% Entradas uniformemente distribuidas na regiao:
% Caso 1 = entre x1=[x1min,x1max] e x2=[x2min,x2max]
% Caso 2 = dentro ou sobre o circulo centrado em
%          Cx=(Cx1, Cx2) e Raio Rx=min(Dx1,Dx2)/3,
%          onde Dx1=x1max-x1min; Dx2=x2max-x2min;
%          Cx1=x1min+Dx1/2; Cx2=x2min+Dx2/2; 
% Caso 3 = entre x1=[x1min,x1max] e x2=[x2min,x2max] mas
%          fora do circulo centrado em (Cx1, Cx2) e Raio Rx
% Caso 4 = dentro ou sobre o triangulo com pontos (x1,x2) =
%          (-1,-1),(1,-1),(0,1) [retas y=kx+1, k=-2 e 2]
% Caso 5 = entre x1=[-1,1] e x2=[-1,1] mas fora do triangulo
%          definido no caso 4
% Caso 6 = como no caso 1 com x1=[-1 1], x2[-1 1], porem nao
%          sao aceitos os pontos no quadrante superior direito
%          (x1=[0 1], x2=[0 1])
% Obs: Nos casos 2 e 3, definiremos Cx como centro do retangulo
%      definido pelos limites em x1 e x2 e Rx=min(Dx1,Dx2)/3
%      onde Dx1=(x1max-x1min), Dx2=(x2max-x2min)
figure(Nf);
Caso=menu('Escolha o caso:', ...
          'Caso 1 = retngulo', ...
          'Caso 2 = crculo', ...
          'Caso 3 = retngulo mas fora do crculo', ...
          'Caso 4 = tringulo', ...
          'Caso 5 = retngulo mas fora do tringulo', ...
          'Caso 6 = em forma de L');
if Caso==4 | Caso==5 | Caso==6;
   x1min=-1; x1max=1; x2min=-1; x2max=1;
end;
Dx1=x1max-x1min; Dx2=x2max-x2min;
Cx1=x1min+Dx1/2; Cx2=x2min+Dx2/2;
Rx=min(Dx1,Dx2)/3;    % Raio do circulo para casos 2 e 3

% Passos do algoritmo de Kohonen:
% Passo 1 = Inicializacao dos pesos
% Passo 2 = Escolha uma entrada x
% Passo 3 = Para cada unidade da rede calcule a distancia D
%           entre x e w => D(i) = ||x-w(i)||
% Passo 4 = Identifique a unidade vencedora v onde
%           D(v) = min[D(i)] para todo i
% Passo 5 = Atualize os pesos das unidades em Viz(v)=vizinhanca
%           da unidade v
%           w(j)=w(j)+neta*[x-w(j)], apenas quando j E Viz(v)
% Passo 6 = diminua neta e o tamanho da vizinhanca Viz(v)

% Para a rede neural com N unidades (distribuidas em
% grade uni-, bi- ou tri-dimensional) a matriz de pesos
% W tem N colunas e Nin linhas, onde Nin=length(x)
% As unidades sao numeradas na sequencia x,y,z
% da esquerda para direita, de baixo para cima,
% da frente para tras

% Caso da rede neural com N unidades distribuidas
% em uma grade bi-dimensional
% N1 unidades no eixo horizontal
% N2 unidades no eixo vertical
% Cxyu = matriz com 2 linhas e N1*N2 colunas
%  o numero da unidade e' dado pelo numero da coluna
%  1a. linha de Cxyu = coordenada x da unidade
%  2a. linha de Cxyu = coordenada y da unidade
Cxyu=zeros([2,N1*N2]);
for ui=1:N1*N2;
     Cxyu(1,ui)=ui-N1*(ceil(ui/N1)-1); % coordenada x
     Cxyu(2,ui)=ceil(ui/N1);           % coordenada y
     %Para teste faca: xui=Cxyu(1,ui); yui=Cxyu(2,ui);
     %text(xui,yui,[num2str(xui) ',' num2str(yui)]);
end;
% Se testando: axis([0 N1+1 0 N2+1]);

%% Passo 1 = Inicializacao dos pesos
W=rand([2,N1*N2]);
W(1,:)=W(1,:)*Dx1+x1min;
W(2,:)=W(2,:)*Dx2+x2min;
% Para teste da rotina de visualizacao dos pesos da rede
% if N1==4 & N2==3;
%   W=[-0.5  0.0  0.5  1.0 -0.5  0.0  0.5 1.0 -0.5  0.0  0.5 1.0;
%      -0.5 -0.5 -0.5 -0.5  0.0  0.0  0.0 0.0  0.5  0.5  0.5 0.5];
% end;

% Visualizacao do estado da rede neural antes do treinamento
% Tracar as linhas horizontais da grade das unidades
figure(Nf); %set(gcf,'Position',[162 86 493 469]);
il=1; % il=numero do segmento de reta
for j=1:N2;
    iu=(j-1)*N1+1:j*N1; % indices das unidades na linha j
    hgr(il)=plot(W(1,iu),W(2,iu));
    il=il+1; hold on;
end;
% Tracar as verticais horizontais da grade das unidades
for i=1:N1;
    iu=i:N1:i+N1*(N2-1); % numero das unidades da coluna i
    hgr(il)=plot(W(1,iu),W(2,iu));
    il=il+1;
end;
% Tracar as marcacoes dos centros das unidades
if N2==1; hgr(il)=plot(W(1,:),W(2,:),'o'); end;

% Mostre a posicao de algumas unidades na grade
figure(Nf); it=0;
if N1>1 & N2>1; uii=[1 N1 N1*(N2-1)+1 N1*N2]; end;
if N2==1; uii=[1 N1]; end; % caso 1 ou 2 dimensoes na entrada
for ui=uii;
    it=it+1;
    hte(it)=text(W(1,ui),W(2,ui),[num2str(ui) ' ']);
    htc(it)=plot(W(1,ui),W(2,ui),'ko');
end;
set(hte,'Color',[0 0 0]); % [1 1 0] = yellow
set(hte,'HorizontalAlignment','right');

% Plotar os limites da regiao dos dados de entrada
% para os casos 2, 3, 4, 5 e 6
% Plotar circulo
if Caso==2 | Caso==3;
   Npc=60; % Numero de segmentos para aproximar o circulo
   fase=[0:1/Npc:1]*(2*pi);
   pcx1=Cx1+Rx*cos(fase);
   pcx2=Cx2+Rx*sin(fase);
   hpr=plot(pcx1,pcx2);
end;
% Plotar o triangulo
if Caso==4 | Caso==5;
   hpr=plot([-1 0 1 -1],[-1 1 -1 -1]);
end;
% Plotar o formato em L
if Caso==6;
   hpr=plot([-1 1 1 0 0 -1 -1],[-1 -1 0 0 1 1 -1]);
end;
if 2 <= Caso & Caso <= 6;
   set(hpr,'Color',[1 0 0]);
   set(hpr,'LineWidth',2);
end;

% Define limites dos eixos da figura
axis([x1min x1max x2min x2max]);  axis('equal');
xlabel('Entrada x1'); ylabel('Entrada x2');
title(['Iterao no. ' num2str(0)]);
drawnow;
disp('Pausing ...'); pause; disp('... out of pause');

%% Treinamento e Visualizacao
t0=clock;
intr=0; % contador do numero de pontos x usados no treinamento
ixv=0;  % contador do numero de pontos x usados entre as visualizacoes
% xtr = matriz com os pontos x usados durante treinamento
% xtr=zeros([size(W,1),Niter*Npiter]);

for iter=1:Niter;
    fprintf('%g ',iter);
    if iter==20*fix(iter/20); fprintf('\n'); end;
    % Ajuste neta e Rv para esta iteracao
    neta=netamin+neta0*(Niter-iter+1)/Niter;
    Rv=Rv0*(Niter-iter+1)/Niter;

    %if iter>50; break; end;
    
    for ipter=1:Npiter;    
        %% Passo 2 = Escolha uma entrada x
        flag0=1;  % flag0 =0 indica que o ponto x deve ser aceito
        while flag0 ~= 0
             x1=x1min+rand*(x1max-x1min);
             x2=x2min+rand*(x2max-x2min);
             x=[x1; x2];
             % Dependendo do caso aceite ou nao este ponto
             if Caso==1; flag0=0; end;
             if Caso==2 | Caso == 3;
                dcx=sqrt((x1-Cx1)^2+(x2-Cx2)^2); % distancia de x para o centro Cx
                if Caso==2 & dcx <= Rx; flag0=0; end;
                if Caso==3 & dcx > Rx ; flag0=0; end;
             end; % "if Caso==2 | Caso == 3"
             if Caso==4 & x2>=-1 & x2<=2*x1+1 & x2<=-2*x1+1;
                flag0=0;
             end;
             if Caso==5 & (x2< -1 | x2> 2*x1+1 | x2> -2*x1+1);
                flag0=0;
             end;
             if Caso==6 & (x1 <= 0 | x2 <= 0);
                flag0=0;
             end;
        end; % "while flag0 ~= 0"
        % Aqui o ponto x e' aceitavel para o caso escolhido
        % Armazene x para posterior plotagem dos pontos de treinamento
        intr=intr+1; ixv=ixv+1;
        %xtr(:,intr)=x;
        %hx(ixv)=plot(x(1),x(2),'+r');
        
        %% Passo 3 = Para cada unidade da rede calcule a distancia D
        %%           entre x e w => D(i) = ||x-w(i)||
        xx=[x(1).*ones(1,N1*N2); x(2).*ones(1,N1*N2)]-W;
        xx=xx.^2; xx1=xx(1,:)+xx(2,:);
        
        %% Passo 4 = Identifique a unidade vencedora uv onde
        %%           D(uv) = min[D(i)] para todo i
        [dmin,uv]=min(xx1);
        xuv=uv-N1*(ceil(uv/N1)-1); % coordenada x de uv
        yuv=ceil(uv/N1);           % coordenada y de uv
                
        %% Passo 5 = Atualize os pesos das unidades em Viz(v)=vizinhanca
        %%           da unidade v
        %%           w(j)=w(j)+neta*[x-w(j)], apenas quando j E Viz(v)
        % Calculo da distancia da unidade ui para a unidade vencedora uv
        for iu=1:N1*N2;
            d2=(xuv-Cxyu(1,iu))^2+(yuv-Cxyu(2,iu))^2;
            hui=exp(-Kexp*d2/(Rv*Rv));
            W(:,iu)=W(:,iu)+neta*hui*(x-W(:,iu));
        end; % "for iu=1:N1*N2"
    
    end; % "for ipter=1:Npiter;"
        
    %% Passo 6 = diminua neta e o tamanho da vizinhanca Viz(v)
    %% Ajuste de neta e tamanho de Viz(v) realizado no inicio do loop    

    % Visualizacao do estado da rede neural durante o treinamento
    if iter==Npv*fix(iter/Npv);
       figure(Nf);
       % remove as linhas e as marcacoes das unidades no desenho anterior
       delete(hgr);
       % Tracar as linhas horizontais da grade das unidades
       strtitle=sprintf('Iterao no. %g',iter);
       title(strtitle);
       il=1; % il=numero do segmento de reta
       for j=1:N2;
           iu=(j-1)*N1+1:j*N1; % indices das unidades na linha j
           hgr(il)=plot(W(1,iu),W(2,iu));
           il=il+1; hold on;
       end; % "for j=1:N2"
       % Tracar as verticais horizontais da grade das unidades
       for i=1:N1;
           iu=i:N1:i+N1*(N2-1); % numero das unidades da coluna i
           hgr(il)=plot(W(1,iu),W(2,iu));
           il=il+1;
       end; % "for i=1:N1"
       % Tracar as marcacoes dos centros das unidades
       if N2==1; hgr(il)=plot(W(1,:),W(2,:),'o'); end;
       % Remover marcacao dos pontos x
       %delete(hx);
       ixv=0;

       % Mostre a posicao de algumas unidades na grade
       delete(hte); delete(htc);
       it=0;
       if N1>1 & N2>1; uii=[1 N1 N1*(N2-1)+1 N1*N2]; end;
       if N2==1; uii=[1 N1]; end;
       for ui=uii;
           it=it+1;
           hte(it)=text(W(1,ui),W(2,ui),[num2str(ui) ' ']);
           htc(it)=plot(W(1,ui),W(2,ui),'ko');
       end;
       set(hte,'Color',[0 0 0]); % [1 1 0] = yellow
       set(hte,'HorizontalAlignment','right');
       axis([x1min x1max x2min x2max]);
       axis('equal'); axis manual;
       drawnow;
       % Visualizacao do efeito da reducao do raio
       for ui=0:max(N1,N2);
           d2=2*ui*ui; % d2=quadrado da diagonal dos catetos ui e ui
           fh(ui+1)=exp(-Kexp*d2/(Rv*Rv));
       end;
       %plot([0:max(N1,N2)],fh); hold on;
       %axis([0 max(N1,N2) 0 1]);
       drawnow;
    end; % "if iter==Npv*fix(iter/Npv)"
    
end; % "for iter=1:Niter"

% EOF koh1.m, 10/Fev/2002

